/*************************************************************************\
* Copyright (c) 2002 The University of Chicago, as Operator of Argonne
* National Laboratory.
* Copyright (c) 2002 The Regents of the University of California, as
* Operator of Los Alamos National Laboratory.
* This file is distributed subject to a Software License Agreement found
* in the file LICENSE that is included with this distribution. 
\*************************************************************************/
#include <stddef.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <jni.h>

#include "dbStatic_DbBase.h"
#include "dbStatic_DbEntry.h"

#include <dbStaticLib.h>
#include <dbBase.h>



#define NBASEINC 10
static DBBASE **papdbBase = NULL;
static long npdbBase = 0;

/*utility routines*/
static void throwException(JNIEnv *env,long status,char *message)
{
    jclass exceptClass;
    char buffer[256];
    char status_buffer[100];
    int rtnval;
    status_buffer[0] = 0;
    rtnval = errSymFind(status,status_buffer);
    if(rtnval) strcpy(status_buffer,"illegal status value");
    strcpy(buffer,message);
    strcat(buffer," ");
    strcat(buffer,status_buffer);
    exceptClass = (*env)->FindClass(env,"dbStatic/DbStaticException");
    (*env)->ThrowNew(env,exceptClass,buffer);
}

static jint getIndexDbBase(JNIEnv *env, jobject obj)
{
    jclass   dbBaseClass;
    jfieldID indexDbBaseID;

    dbBaseClass = (*env)->GetObjectClass(env,obj);
    indexDbBaseID = (*env)->GetFieldID(env,dbBaseClass,"indexDbBase",
	"I");
    return((*env)->GetIntField(env,obj,indexDbBaseID));
}

static void putIndexDbBase(JNIEnv *env, jobject obj, jint indexDbBase)
{
    jclass   dbBaseClass;
    jfieldID indexDbBaseID;

    dbBaseClass = (*env)->GetObjectClass(env,obj);
    indexDbBaseID = (*env)->GetFieldID(env,dbBaseClass,"indexDbBase",
	"I");
    (*env)->SetIntField(env,obj,indexDbBaseID,indexDbBase);
}

/*native method implementation*/
JNIEXPORT void JNICALL Java_dbStatic_DbBase_alloc (JNIEnv *env, jobject obj)
{
    int i;
    jint indexDbBase = -1;

    for(i=0; i<npdbBase; i++) {
	if(!papdbBase[i]) {
	   indexDbBase = i;
	   break;
	}
    }
    if(indexDbBase==-1) {
	DBBASE **paptemp;

	paptemp = dbCalloc(npdbBase+NBASEINC,sizeof(DBBASE *));
	for(i=0; i<npdbBase; i++) paptemp[i] = papdbBase[i];
	if(npdbBase) free((void *)papdbBase);
	papdbBase = paptemp;
	indexDbBase = npdbBase;
	npdbBase += NBASEINC;
    }
    papdbBase[indexDbBase] = dbAllocBase();
    putIndexDbBase(env,obj,indexDbBase);
}

JNIEXPORT void JNICALL Java_dbStatic_DbBase_free(JNIEnv *env, jobject obj)
{
    jint indexDbBase = -1;
    
    indexDbBase = getIndexDbBase(env,obj);
    dbFreeBase(papdbBase[indexDbBase]);
    papdbBase[indexDbBase] = NULL;
}

JNIEXPORT void JNICALL Java_dbStatic_DbBase_readDatabase( JNIEnv *env,
    jobject obj, jstring filename, jstring path, jstring substitutions)
{
    jint indexDbBase = -1;
    const char *str;
    DBBASE     *pdbbase;
    long       status;
    
    indexDbBase = getIndexDbBase(env,obj);
    pdbbase = papdbBase[indexDbBase];
    str = (*env)->GetStringUTFChars(env, filename, 0);
    status = dbReadDatabase(&pdbbase,str,NULL,NULL);

    if(status) throwException(env,status,"From dbReadDatabase");
    (*env)->ReleaseStringUTFChars(env, filename, str);
}

JNIEXPORT void JNICALL Java_dbStatic_DbBase_path
  (JNIEnv *env, jobject obj, jstring path)
{
    jint indexDbBase = -1;
    const char *str;
    DBBASE     *pdbbase;
    long       status;
 
    indexDbBase = getIndexDbBase(env,obj);
    pdbbase = papdbBase[indexDbBase];
    str = (*env)->GetStringUTFChars(env, path, 0);
    status = dbPath(pdbbase,str);
 
    if(status) throwException(env,status,"From dbReadDatabase");
    (*env)->ReleaseStringUTFChars(env, path, str);
}

JNIEXPORT void JNICALL Java_dbStatic_DbBase_addPath
  (JNIEnv *env, jobject obj, jstring path)
{
    jint indexDbBase = -1;
    const char *str;
    DBBASE     *pdbbase;
    long       status;
 
    indexDbBase = getIndexDbBase(env,obj);
    pdbbase = papdbBase[indexDbBase];
    str = (*env)->GetStringUTFChars(env, path, 0);
    status = dbAddPath(pdbbase,str);
 
    if(status) throwException(env,status,"From dbReadDatabase");
    (*env)->ReleaseStringUTFChars(env, path, str);
}

JNIEXPORT void JNICALL Java_dbStatic_DbBase_writeRecord( JNIEnv *env,
    jobject obj, jstring filename, jstring recordTypename, jint level)
{
    jint indexDbBase = -1;
    const char *str;
    const char *recType;
    DBBASE     *pdbbase;
    long       status;
 
    indexDbBase = getIndexDbBase(env,obj);
    pdbbase = papdbBase[indexDbBase];
    str = (*env)->GetStringUTFChars(env, filename, 0);
    recType = (*env)->GetStringUTFChars(env, recordTypename,0);
   if (strlen(recType)==0)   
 /*    if ( (recType == NULL) || strlen(recType)==0) */ 
       status = dbWriteRecord(pdbbase, str, NULL, level);
    else
       status = dbWriteRecord(pdbbase,str, recType,level);
    if(status) throwException(env,status,"From dbWriteRecord");
    (*env)->ReleaseStringUTFChars(env, filename, str);
    (*env)->ReleaseStringUTFChars(env, recordTypename, recType);
}

JNIEXPORT void JNICALL Java_dbStatic_DbBase_writeMenu
  (JNIEnv *env, jobject obj, jstring filename, jstring menuName)
{
    jint indexDbBase = -1;
    const char *strfile;
    const char *strmenu;
    DBBASE     *pdbbase;
    long       status;
 
    indexDbBase = getIndexDbBase(env,obj);
    pdbbase = papdbBase[indexDbBase];
    strfile = (*env)->GetStringUTFChars(env, filename, 0);
    strmenu = (*env)->GetStringUTFChars(env, menuName, 0); 
    status = dbWriteMenu(pdbbase,(char *)strfile, (char *)strmenu);
 
    if(status) throwException(env,status,"From dbWriteMenu");
    (*env)->ReleaseStringUTFChars(env, filename, strfile);
    (*env)->ReleaseStringUTFChars(env, menuName, strmenu);
}

JNIEXPORT void JNICALL Java_dbStatic_DbBase_writeRecordType
  (JNIEnv *env, jobject obj, jstring filename, jstring recordTypeName)
{
    jint indexDbBase = -1;
    const char *strfile;
    const char *strrecord;
    DBBASE     *pdbbase;
    long       status;
 
    indexDbBase = getIndexDbBase(env,obj);
    pdbbase = papdbBase[indexDbBase];
    strfile = (*env)->GetStringUTFChars(env, filename, 0);
    strrecord = (*env)->GetStringUTFChars(env, recordTypeName, 0); 
    status = dbWriteRecordType(pdbbase,(char *)strfile, (char *)strrecord);
 
    if(status) throwException(env,status,"From dbWriteRecordTypeu");
    (*env)->ReleaseStringUTFChars(env, filename, strfile);
    (*env)->ReleaseStringUTFChars(env, recordTypeName, strrecord);
}

JNIEXPORT void JNICALL Java_dbStatic_DbBase_writeDevice
  (JNIEnv *env, jobject obj, jstring filename)
{
    jint indexDbBase = -1;
    const char *strfile;
    DBBASE     *pdbbase;
    long       status;
 
    indexDbBase = getIndexDbBase(env,obj);
    pdbbase = papdbBase[indexDbBase];
    strfile = (*env)->GetStringUTFChars(env, filename, 0);
    status = dbWriteDevice(pdbbase,strfile);
    if(status) throwException(env,status,"From dbWriteDevice");
    (*env)->ReleaseStringUTFChars(env, filename, strfile);
}

JNIEXPORT void JNICALL Java_dbStatic_DbBase_writeDriver
  (JNIEnv *env, jobject obj, jstring filename)
{
    jint indexDbBase = -1;
    const char *strfile;
    DBBASE     *pdbbase;
    long       status;
 
    indexDbBase = getIndexDbBase(env,obj);
    pdbbase = papdbBase[indexDbBase];
    strfile = (*env)->GetStringUTFChars(env, filename, 0);
    status = dbWriteDriver(pdbbase,strfile);
    if(status) throwException(env,status,"From dbWriteDriver");
    (*env)->ReleaseStringUTFChars(env, filename, strfile);
}

JNIEXPORT void JNICALL Java_dbStatic_DbBase_writeBreaktable
  (JNIEnv *env, jobject obj, jstring filename)
{
    jint indexDbBase = -1;
    const char *strfile;
    DBBASE     *pdbbase;
    long       status;
 
    indexDbBase = getIndexDbBase(env,obj);
    pdbbase = papdbBase[indexDbBase];
    strfile = (*env)->GetStringUTFChars(env, filename, 0);
    status = dbWriteBreaktable(pdbbase,strfile);
    if(status) throwException(env,status,"From dbWriteBreaktabler");
    (*env)->ReleaseStringUTFChars(env, filename, strfile);
}


#define NENTRYINC 10
static DBENTRY **papdbEntry = NULL;
static long npdbEntry = 0;

static jint getIndexDbEntry(JNIEnv *env, jobject obj)
{
    jclass   dbEntryClass;
    jfieldID indexDbEntryID;

    dbEntryClass = (*env)->GetObjectClass(env,obj);
    indexDbEntryID = (*env)->GetFieldID(env,dbEntryClass,"indexDbEntry",
	"I");
    return((*env)->GetIntField(env,obj,indexDbEntryID));
}

static void putIndexDbEntry(JNIEnv *env, jobject obj, jint indexDbEntry)
{
    jclass   dbEntryClass;
    jfieldID indexDbEntryID;

    dbEntryClass = (*env)->GetObjectClass(env,obj);
    indexDbEntryID = (*env)->GetFieldID(env,dbEntryClass,"indexDbEntry",
	"I");
    (*env)->SetIntField(env,obj,indexDbEntryID,indexDbEntry);
}

static jint getAllocFormNstrings
  (JNIEnv *env, jobject obj)
{
     jclass     dbEntryClass;
     jfieldID   allocFormNstringsID;

     dbEntryClass = (*env)->GetObjectClass(env, obj);
     allocFormNstringsID = (*env)->GetFieldID(env, dbEntryClass,
"allocFormNstrings", "I");
     return((*env)->GetIntField(env,obj,allocFormNstringsID));
}

static void putAllocFormNstrings
  (JNIEnv *env, jobject obj, jint allocFormNstrings)
{
     jclass   dbEntryClass;
     jfieldID allocFormNstringsID;

     dbEntryClass = (*env)->GetObjectClass(env, obj);
     allocFormNstringsID = (*env)->GetFieldID(env, dbEntryClass,
"allocFormNstrings", "I");
     (*env)->SetIntField(env, obj, allocFormNstringsID,
allocFormNstrings);
}

JNIEXPORT void JNICALL Java_dbStatic_DbEntry_alloc
    (JNIEnv *env, jobject obj,jint indexDbBase)
{
    int i;
    jint indexDbEntry = -1;

    for(i=0; i<npdbEntry; i++) {
	if(!papdbEntry[i]) {
	   indexDbEntry = i;
	   break;
	}
    }
    if(indexDbEntry==-1) {
	DBENTRY **paptemp;
	paptemp = dbCalloc(npdbEntry+NENTRYINC,sizeof(DBENTRY *));
	for(i=0; i<npdbEntry; i++) paptemp[i] = papdbEntry[i];
	if(npdbEntry) free((void *)papdbEntry);
	papdbEntry = paptemp;
	indexDbEntry = npdbEntry;
	npdbEntry += NENTRYINC;
    }
    papdbEntry[indexDbEntry] = dbAllocEntry(papdbBase[indexDbBase]);
    putIndexDbEntry(env,obj,indexDbEntry);
    putAllocFormNstrings(env, obj, -1);
}

JNIEXPORT void JNICALL Java_dbStatic_DbEntry_free(JNIEnv *env, jobject obj)
{
    jint indexDbEntry = -1;
    
    indexDbEntry = getIndexDbEntry(env,obj);
    dbFreeEntry(papdbEntry[indexDbEntry]);
    papdbEntry[indexDbEntry] = NULL;
}

JNIEXPORT jboolean JNICALL Java_dbStatic_DbEntry_findRecordType
    (JNIEnv *env, jobject obj, jstring recordTypeName)
{
    jint       indexDbEntry = -1;
    const char *str;
    long       status;
    
    indexDbEntry = getIndexDbEntry(env,obj);
    str = (*env)->GetStringUTFChars(env, recordTypeName, 0);
    status = dbFindRecordType(papdbEntry[indexDbEntry],(char *)str);
    (*env)->ReleaseStringUTFChars(env, recordTypeName, str);
    return(!status);
}

JNIEXPORT jboolean JNICALL Java_dbStatic_DbEntry_firstRecordType
    (JNIEnv *env, jobject obj)
{
    jint       indexDbEntry = -1;
    long       status;
    
    indexDbEntry = getIndexDbEntry(env,obj);
    status = dbFirstRecordType(papdbEntry[indexDbEntry]);
    return(!status);
}

JNIEXPORT jboolean JNICALL Java_dbStatic_DbEntry_nextRecordType
    (JNIEnv *env, jobject obj)
{
    jint       indexDbEntry = -1;
    long       status;
    
    indexDbEntry = getIndexDbEntry(env,obj);
    status = dbNextRecordType(papdbEntry[indexDbEntry]);
    return(!status);
}

JNIEXPORT jstring JNICALL Java_dbStatic_DbEntry_getRecordTypeName
    (JNIEnv *env, jobject obj)
{
    jint       indexDbEntry = -1;
    long       status;
    jobject    recordTypeName;
    char       *str;
    
    indexDbEntry = getIndexDbEntry(env,obj);
    str = dbGetRecordTypeName(papdbEntry[indexDbEntry]);
    if(str==NULL) throwException(env,-1,"From dbGetRecordTypeName");
    recordTypeName = (*env)->NewStringUTF(env,str);
    return(recordTypeName);
}

JNIEXPORT jint JNICALL Java_dbStatic_DbEntry_getNRecordTypes
    (JNIEnv *env, jobject obj)
{
    jint       indexDbEntry = -1;
    int        n;
    
    indexDbEntry = getIndexDbEntry(env,obj);
    n = dbGetNRecordTypes(papdbEntry[indexDbEntry]);
    return((jint)n);
}

JNIEXPORT jboolean JNICALL Java_dbStatic_DbEntry_firstField
  (JNIEnv *env, jobject obj, jboolean dctonly)
{
    jint       indexDbEntry = -1;
    long       status;
 
    indexDbEntry = getIndexDbEntry(env,obj);
    status = dbFirstField(papdbEntry[indexDbEntry],dctonly);
    return(!status);
}

JNIEXPORT jboolean JNICALL Java_dbStatic_DbEntry_nextField
  (JNIEnv *env, jobject obj, jboolean dctonly)
{
    jint       indexDbEntry = -1;
    long       status;
 
    indexDbEntry = getIndexDbEntry(env,obj);
    status = dbNextField(papdbEntry[indexDbEntry],dctonly);
    return(!status);
}

JNIEXPORT jint JNICALL Java_dbStatic_DbEntry_getFieldType
  (JNIEnv *env, jobject obj)
{
    jint       indexDbEntry = -1;
    int        type;
 
    indexDbEntry = getIndexDbEntry(env,obj);
    type = dbGetFieldType(papdbEntry[indexDbEntry]);
    return((jint)type);
}

JNIEXPORT jint JNICALL Java_dbStatic_DbEntry_getNFields
  (JNIEnv *env, jobject obj, jboolean dctonly)
{
    jint       indexDbEntry = -1;
    int        type;
 
    indexDbEntry = getIndexDbEntry(env,obj);
    type = dbGetNFields(papdbEntry[indexDbEntry],dctonly);
    return((jint)type);
}

JNIEXPORT jstring JNICALL Java_dbStatic_DbEntry_getFieldName
  (JNIEnv *env, jobject obj)
{
    jint       indexDbEntry = -1;
    long       status;
    jobject    fieldName;
    char       *str;
 
    indexDbEntry = getIndexDbEntry(env,obj);
    str = dbGetFieldName(papdbEntry[indexDbEntry]);
    if(str==NULL) throwException(env,-1,"From dbGetFieldName");
    fieldName = (*env)->NewStringUTF(env,str);
    return(fieldName);
}

JNIEXPORT jstring JNICALL Java_dbStatic_DbEntry_getDefault
  (JNIEnv *env, jobject obj)
{
    jint       indexDbEntry = -1;
    long       status;
    jobject    value;
    char       *str;
 
    indexDbEntry = getIndexDbEntry(env,obj);
    str = dbGetDefault(papdbEntry[indexDbEntry]);
    if(str==NULL) throwException(env,-1,"From dbGetDefault");
    value = (*env)->NewStringUTF(env,str);
    return(value);
}

JNIEXPORT jstring JNICALL Java_dbStatic_DbEntry_getPrompt
  (JNIEnv *env, jobject obj)
{
    jint       indexDbEntry = -1;
    long       status;
    jobject    value;
    char       *str;
 
    indexDbEntry = getIndexDbEntry(env,obj);
    str = dbGetPrompt(papdbEntry[indexDbEntry]);
    if(str==NULL) throwException(env,-1,"From dbGetPrompt");
    value = (*env)->NewStringUTF(env,str);
    return(value);
}

JNIEXPORT jboolean JNICALL Java_dbStatic_DbEntry_isDctField
  (JNIEnv *env, jobject obj)
{
    jint       indexDbEntry = -1;
    int        group;
 
    indexDbEntry = getIndexDbEntry(env,obj);
    group = dbGetPromptGroup(papdbEntry[indexDbEntry]);
    return((group>0)?FALSE:TRUE);
}

JNIEXPORT void JNICALL Java_dbStatic_DbEntry_createRecord
  (JNIEnv *env, jobject obj, jstring recordName)
{
    jint       indexDbEntry = -1;
    const char *str;
    long       status;
    
    indexDbEntry = getIndexDbEntry(env,obj);
    str = (*env)->GetStringUTFChars(env, recordName, 0);
    status = dbCreateRecord(papdbEntry[indexDbEntry],(char *)str);
    if(status) throwException(env,status,"From dbCreateRecord");
    (*env)->ReleaseStringUTFChars(env, recordName, str);
}

JNIEXPORT void JNICALL Java_dbStatic_DbEntry_deleteRecord
  (JNIEnv *env, jobject obj)
{
    jint       indexDbEntry = -1;
    long       status;
    
    indexDbEntry = getIndexDbEntry(env,obj);
    status = dbDeleteRecord(papdbEntry[indexDbEntry]);
    if(status) throwException(env,status,"From dbDeleteRecord");
}

JNIEXPORT void JNICALL Java_dbStatic_DbEntry_freeRecords
  (JNIEnv *env, jobject obj)
{
    jint indexDbBase = -1;
    DBBASE     *pdbbase;
    long       status;
 
    indexDbBase = getIndexDbBase(env,obj);
    pdbbase = papdbBase[indexDbBase];
    status = dbFreeRecords(pdbbase);
    if(status) throwException(env,status,"From dbFreeRecords");
}

JNIEXPORT jboolean JNICALL Java_dbStatic_DbEntry_findRecord
  (JNIEnv *env, jobject obj, jstring recordName)
{
    jint       indexDbEntry = -1;
    const char *str;
    long       status;
    
    indexDbEntry = getIndexDbEntry(env,obj);
    str = (*env)->GetStringUTFChars(env, recordName, 0);
    status = dbFindRecord(papdbEntry[indexDbEntry],(char *)str);
    (*env)->ReleaseStringUTFChars(env, recordName, str);
    return(!status);
}

JNIEXPORT jboolean JNICALL Java_dbStatic_DbEntry_firstRecord
  (JNIEnv *env, jobject obj)
{
    jint       indexDbEntry = -1;
    long       status;
    
    indexDbEntry = getIndexDbEntry(env,obj);
    status = dbFirstRecord(papdbEntry[indexDbEntry]);
    return(!status);
}

JNIEXPORT jboolean JNICALL Java_dbStatic_DbEntry_nextRecord
  (JNIEnv *env, jobject obj)
{
    jint       indexDbEntry = -1;
    long       status;
    
    indexDbEntry = getIndexDbEntry(env,obj);
    status = dbNextRecord(papdbEntry[indexDbEntry]);
    return(!status);
}

JNIEXPORT jint JNICALL Java_dbStatic_DbEntry_getNRecords
  (JNIEnv *env, jobject obj)
{
    jint       indexDbEntry = -1;
    int        n;
 
    indexDbEntry = getIndexDbEntry(env,obj);
    n = dbGetNRecords(papdbEntry[indexDbEntry]);
    return((jint)n);
}

JNIEXPORT jstring JNICALL Java_dbStatic_DbEntry_getRecordName
  (JNIEnv *env, jobject obj)
{
    jint       indexDbEntry = -1;
    long       status;
    jobject    recordName;
    char       *str;
   
    indexDbEntry = getIndexDbEntry(env,obj);
    str = dbGetRecordName(papdbEntry[indexDbEntry]);
    if(str==NULL) throwException(env,-1,"From dbGetRecordName");
    recordName = (*env)->NewStringUTF(env,str);
    return(recordName);
}

JNIEXPORT void JNICALL Java_dbStatic_DbEntry_renameRecord
  (JNIEnv *env, jobject obj, jstring newName)
{
    jint       indexDbEntry = -1;
    const char *str;
    long       status;
 
    indexDbEntry = getIndexDbEntry(env,obj);
    str = (*env)->GetStringUTFChars(env, newName, 0);
    status = dbRenameRecord(papdbEntry[indexDbEntry],(char *)str);
    if(status) throwException(env,status,"From dbRenameRecord");
    (*env)->ReleaseStringUTFChars(env, newName, str);
}

JNIEXPORT void JNICALL Java_dbStatic_DbEntry_copyRecord
  (JNIEnv *env, jobject obj, jstring newName, jboolean overWriteOK)
{
    jint       indexDbEntry = -1;
    const char *str;
    long       status;

    indexDbEntry = getIndexDbEntry(env,obj);
    str = (*env)->GetStringUTFChars(env, newName, 0);
     
    status = dbCopyRecord(papdbEntry[indexDbEntry],(char *)str,overWriteOK);
    if(status) throwException(env,status,"From dbCopyRecord");
    (*env)->ReleaseStringUTFChars(env, newName, str);
}

JNIEXPORT void JNICALL Java_dbStatic_DbEntry_setVisibility
  (JNIEnv *env, jobject obj, jboolean makeVisible)
{
    jint       indexDbEntry = -1;
    long       status;
 
    indexDbEntry = getIndexDbEntry(env,obj);
    if(makeVisible) {
        status = dbVisibleRecord(papdbEntry[indexDbEntry]);
        if(status) throwException(env,status,"From dbVisibleRecord");
    } else {
        status = dbInvisibleRecord(papdbEntry[indexDbEntry]);
        if(status) throwException(env,status,"From dbInvisibleRecord");
    }
}

JNIEXPORT jboolean JNICALL Java_dbStatic_DbEntry_getVisibility
  (JNIEnv *env, jobject obj)
{
    jint       indexDbEntry = -1;
    int        result;
    indexDbEntry = getIndexDbEntry(env,obj); 
    return(dbIsVisibleRecord(papdbEntry[indexDbEntry])); 
}

JNIEXPORT jboolean JNICALL Java_dbStatic_DbEntry_findField
  (JNIEnv *env, jobject obj, jstring fieldName)
{
    jint       indexDbEntry = -1;
    const char *str;
    long       status;
 
    indexDbEntry = getIndexDbEntry(env,obj);
    str = (*env)->GetStringUTFChars(env, fieldName, 0);
    status = dbFindField(papdbEntry[indexDbEntry],(char *)str);
    (*env)->ReleaseStringUTFChars(env, fieldName, str);
    return(!status);
}

JNIEXPORT jboolean JNICALL Java_dbStatic_DbEntry_foundField
  (JNIEnv *env, jobject obj)
{
    jint       indexDbEntry = -1;
    long       status;
 
    indexDbEntry = getIndexDbEntry(env,obj);
    return(dbFoundField(papdbEntry[indexDbEntry]));
}

JNIEXPORT jstring JNICALL Java_dbStatic_DbEntry_getString
  (JNIEnv *env, jobject obj)
{
    jint       indexDbEntry = -1;
    long       status;
    jobject    value;
    char       *str;
 
    indexDbEntry = getIndexDbEntry(env,obj);
    str = dbGetString(papdbEntry[indexDbEntry]);
    value = (*env)->NewStringUTF(env,str);
    return(value);
}

JNIEXPORT void JNICALL Java_dbStatic_DbEntry_putString
  (JNIEnv *env, jobject obj, jstring value)
{
    jint       indexDbEntry = -1;
    const char *str;
    long       status;
 
    indexDbEntry = getIndexDbEntry(env,obj);
    str = (*env)->GetStringUTFChars(env, value, 0);
    status = dbPutString(papdbEntry[indexDbEntry],(char *)str);
    if(status) throwException(env,status,"From dbPutString");
    (*env)->ReleaseStringUTFChars(env, value, str);
}

JNIEXPORT jstring JNICALL Java_dbStatic_DbEntry_verify
  (JNIEnv *env, jobject obj, jstring value)
{
    jint       indexDbEntry = -1;
    const char *str;
    char       *message;
    jobject    rtnmessage;
 
    indexDbEntry = getIndexDbEntry(env,obj);
    str = (*env)->GetStringUTFChars(env, value, 0);
    message = dbVerify(papdbEntry[indexDbEntry],(char *)str);
    (*env)->ReleaseStringUTFChars(env, value, str);
    if((message==NULL) || (strcmp(message,"") ==0)) {
       return(NULL);
     }

    rtnmessage =  (*env)->NewStringUTF(env,message);
    return(rtnmessage);
}

JNIEXPORT jstring JNICALL Java_dbStatic_DbEntry_getRange
  (JNIEnv *env, jobject obj)

{
    jint       indexDbEntry = -1;
    jobject    value;
    char       *str;
 
    indexDbEntry = getIndexDbEntry(env,obj);
    str = dbGetRange(papdbEntry[indexDbEntry]);
    value = (*env)->NewStringUTF(env,str);
    return(value);
}

JNIEXPORT jboolean JNICALL Java_dbStatic_DbEntry_isDefaultValue
  (JNIEnv *env, jobject obj)
{
    jint       indexDbEntry = -1;
    int        result;
 
    indexDbEntry = getIndexDbEntry(env,obj);
    return(dbIsDefaultValue(papdbEntry[indexDbEntry]));
}

JNIEXPORT jint JNICALL Java_dbStatic_DbEntry_getNMenuChoices
  (JNIEnv *env, jobject obj)
{
    jint       indexDbEntry = -1;
    int        n;
 
    indexDbEntry = getIndexDbEntry(env,obj);
    n = dbGetNMenuChoices(papdbEntry[indexDbEntry]);
    return((jint)n);
}

JNIEXPORT jint JNICALL Java_dbStatic_DbEntry_getMenuIndex
  (JNIEnv *env, jobject obj)
{
    jint       indexDbEntry = -1;
    int        index;
    indexDbEntry = getIndexDbEntry(env,obj); 
    index = dbGetMenuIndex(papdbEntry[indexDbEntry]); 
    return((jint)index);
}

JNIEXPORT void JNICALL Java_dbStatic_DbEntry_putMenuIndex 
   (JNIEnv *env, jobject obj, jint index)
{ 
    jint       indexDbEntry = -1;
    long       status;
    indexDbEntry = getIndexDbEntry(env,obj);
    status = dbPutMenuIndex(papdbEntry[indexDbEntry],index);
    if(status) throwException(env,status,"From dbPutMenuIndex");
}

JNIEXPORT jstring JNICALL Java_dbStatic_DbEntry_getMenuStringFromIndex
  (JNIEnv *env, jobject obj, jint index)
{
    jint       indexDbEntry = -1;
    long       status;
    jobject    value;
    char       *str;
 
    indexDbEntry = getIndexDbEntry(env,obj);
    str = dbGetMenuStringFromIndex(papdbEntry[indexDbEntry],index);
    value = (*env)->NewStringUTF(env,str);
    return(value);
}

JNIEXPORT jint JNICALL Java_dbStatic_DbEntry_getMenuIndexFromString
  (JNIEnv *env, jobject obj, jstring value)
{
    jint       indexDbEntry = -1;
    const char *str;
    int        index;
 
    indexDbEntry = getIndexDbEntry(env,obj);
    str = (*env)->GetStringUTFChars(env, value, 0);
    index = dbGetMenuIndexFromString(papdbEntry[indexDbEntry],(char *)str);
    (*env)->ReleaseStringUTFChars(env, value, str);
}

JNIEXPORT jint JNICALL Java_dbStatic_DbEntry_allocForm
  (JNIEnv *env, jobject obj)
{
     jint    indexDbEntry = -1;
     int     n;
     jint     allocFormNstrings;
 
     indexDbEntry = getIndexDbEntry(env, obj);
     n = dbAllocForm(papdbEntry[indexDbEntry]);
     allocFormNstrings = getAllocFormNstrings(env,obj);
     putAllocFormNstrings(env,obj,(jint)n);
     return((jint)n);
}
 
JNIEXPORT jboolean JNICALL Java_dbStatic_DbEntry_freeForm 
  (JNIEnv *env, jobject obj) 
{    
     jint    indexDbEntry = -1; 
     long     status;
     indexDbEntry = getIndexDbEntry(env, obj);
     status = dbFreeForm(papdbEntry[indexDbEntry]);
     putAllocFormNstrings(env,obj,-1);
     return(!status);
} 

JNIEXPORT jobjectArray JNICALL Java_dbStatic_DbEntry_getFormPrompt
  (JNIEnv *env, jobject obj)
{
     jint     indexDbEntry = -1;
     char     **str;
     int      num, i; 
     jobjectArray  newArr;
     jstring       utf_str;
     jclass    clazz;
     jint     allocFormNstrings;
    
     clazz = (*env)->FindClass(env,"java/lang/String");
     indexDbEntry = getIndexDbEntry(env, obj);
     allocFormNstrings = getAllocFormNstrings(env,obj);
     num = allocFormNstrings;
     if(num==-1) {
	 throwException(env,-1,"From dbGetFormPrompt --  call dbAllocForm before ");
         return(NULL);
       } 
      str = dbGetFormPrompt(papdbEntry[indexDbEntry]);
      newArr = (*env)->NewObjectArray(env, (jint)num, clazz, NULL);	
	for(i=0; i<num; i++) { 
	  utf_str = (*env)->NewStringUTF(env,str[i]);     
   	   (*env)->SetObjectArrayElement(env,newArr, i, utf_str);

   	   (*env)->DeleteLocalRef(env,utf_str);  
         }
        return (newArr);
}

JNIEXPORT jobjectArray JNICALL Java_dbStatic_DbEntry_getFormValue 
  (JNIEnv *env, jobject obj)
{
    jint         indexDbEntry = -1; 
    char         **str; 
    int          num, i; 
    jstring      utf_str;
    jobjectArray  newArr; 
    jclass        clazz;
    jint         allocFormNstrings;


    clazz = (*env)->FindClass(env, "java/lang/String");
    indexDbEntry = getIndexDbEntry(env, obj);
        allocFormNstrings = getAllocFormNstrings(env, obj);	
        num= allocFormNstrings;
        if (num == -1) {
	   throwException(env, -1, "From dbGetFormValue");
  	   return(NULL);
	}
	newArr = (*env)->NewObjectArray(env, allocFormNstrings, clazz, NULL);
        str = dbGetFormValue(papdbEntry[indexDbEntry]);
 
	for(i=0; i<num; i++) { 
           utf_str = (*env)->NewStringUTF(env,*str); 
          (*env)->SetObjectArrayElement(env,newArr, i, utf_str);     
           (*env)->DeleteLocalRef(env,utf_str);    
           str++;   
        }
     return (newArr);
}

JNIEXPORT jboolean JNICALL Java_dbStatic_DbEntry_putForm
  (JNIEnv *env, jobject obj, jobjectArray value)
{    
     jint       indexDbEntry = -1; 
     const char **str;
     int        i;
     long       status;
     jsize      len;
     jobject    body;
     jstring    j_str;

     indexDbEntry = getIndexDbEntry(env, obj);
     len = (*env)->GetArrayLength(env,value);
     str = (char *) calloc(len, sizeof(char *)); 
     for(i=0; i<len; i++) { 
	j_str = (*env)->GetObjectArrayElement(env,value, i);
	str[i] = (*env)->GetStringUTFChars(env, j_str, 0);
     }

     status = dbPutForm(papdbEntry[indexDbEntry],(char **)str);

     for(i=0; i<len; i++){
         (*env)->ReleaseStringUTFChars(env, j_str, str[i]);
     }

     return(!status); 
}

JNIEXPORT jboolean JNICALL Java_dbStatic_DbEntry_verifyForm 
(JNIEnv *env, jobject obj, jobjectArray value, jobjectArray error) 
{
     jint       indexDbEntry = -1;
     const char       **str; 
     int        i, num, m;
     char       **vstr;
     jsize      len;
     jstring    j_str, utf_str; 
     jobjectArray    newArr;
     jclass          clazz; 
     jint        allocFormNstrings;
     long    status;   

     indexDbEntry = getIndexDbEntry(env, obj);
     len = (*env)->GetArrayLength(env, value);
     str = (char *) calloc(len, sizeof(char *)); 
     for(i=0; i<len; i++){
         j_str = (*env)->GetObjectArrayElement(env, value, i);
         str[i] = (*env)->GetStringUTFChars(env, j_str, 0);
     }
     allocFormNstrings = getAllocFormNstrings(env, obj);
     num = allocFormNstrings;
      if(num==-1) {
	  throwException(env, -1, "From dbVerifyForm");
	  return (FALSE);
      } 

     vstr = dbVerifyForm(papdbEntry[indexDbEntry],(char **)str); 
     for(i=0; i<len; i++){
         (*env)->ReleaseStringUTFChars(env, j_str, str[i]);
     }
      if  (vstr == NULL)   {
          return(FALSE); 
      }
   
      for(i=0;i<num;i++) {
   	utf_str = (*env)->NewStringUTF(env,vstr[i]);
	(*env)->SetObjectArrayElement(env, error, i, utf_str);
	(*env)->DeleteLocalRef(env,utf_str);
      }
      return(TRUE);
}

JNIEXPORT jstring JNICALL Java_dbStatic_DbEntry_getRelatedField
  (JNIEnv *env	, jobject obj)
{
     jint    indexDbEntry = -1;
     jobject value=NULL;
     char    *str;

     indexDbEntry = getIndexDbEntry(env,obj);
     str = dbGetRelatedField(papdbEntry[indexDbEntry]);
     if(!str) throwException(env,-1,"From dbGetRelatedField");
     else value = (*env)->NewStringUTF(env, str);
     return(value);
}

JNIEXPORT jint JNICALL Java_dbStatic_DbEntry_getNLinks
  (JNIEnv *env, jobject obj)
{
 
     jint     indexDbEntry = -1;
     int      index;
 
     indexDbEntry = getIndexDbEntry(env, obj);
     index = dbGetNLinks(papdbEntry[indexDbEntry]);
     return((jint)index);
}
 
JNIEXPORT jboolean JNICALL Java_dbStatic_DbEntry_getLinkField
  (JNIEnv *env, jobject obj, jint index)
{
     jint indexDbEntry = -1;
     long status;
 
     indexDbEntry = getIndexDbEntry(env, obj);
     status = dbGetLinkField(papdbEntry[indexDbEntry], index);
     return(!status);
}
 
JNIEXPORT jint JNICALL Java_dbStatic_DbEntry_getLinkType
  (JNIEnv *env, jobject obj)
{ 
     jint     indexDbEntry = -1; 
     int      n;
     indexDbEntry = getIndexDbEntry(env, obj);
     n = dbGetLinkType(papdbEntry[indexDbEntry]);
     return((jint)n);
}
 
JNIEXPORT jboolean JNICALL Java_dbStatic_DbEntry_cvtLinkToConstant
  (JNIEnv *env, jobject obj)
{
     jint indexDbEntry = -1;
     long status;
 
     indexDbEntry = getIndexDbEntry(env, obj);
     status = dbCvtLinkToConstant(papdbEntry[indexDbEntry]);
     return(!status);
}

JNIEXPORT jboolean JNICALL Java_dbStatic_DbEntry_cvtLinkToPvlink
  (JNIEnv *env, jobject obj)
{
     jint indexDbEntry = -1;
     long status; 
     indexDbEntry = getIndexDbEntry(env, obj); 
     status = dbCvtLinkToPvlink(papdbEntry[indexDbEntry]);
     return(!status);
}

JNIEXPORT void JNICALL Java_dbStatic_DbBase_dumpPath (JNIEnv *env, jobject obj)
{
    jint    indexDbBase = -1;
    const char *str;
    DBBASE *pdbbase;

   indexDbBase = getIndexDbBase(env, obj);
   pdbbase = papdbBase[indexDbBase];
   dbDumpPath(pdbbase);
}

JNIEXPORT void JNICALL Java_dbStatic_DbBase_dumpRecord
  (JNIEnv *env, jobject obj, jstring precordTypeName, jint level) 
{     
     jint     indexDbBase = -1;
     const char *str; DBBASE *pdbbase; const char *tstr;

    indexDbBase = getIndexDbBase(env, obj);
    pdbbase = papdbBase[indexDbBase];
    str = (*env)->GetStringUTFChars(env, precordTypeName,0);
    tstr = ((str==0) || (strlen(str)==0)) ? 0 : str;
 
    dbDumpRecord(pdbbase,(char *)tstr, level);
} 

JNIEXPORT void JNICALL Java_dbStatic_DbBase_dumpMenu
  (JNIEnv *env, jobject obj, jstring menuName) 
{
     jint    indexDbBase = -1;
     const   char *str;
     DBBASE  *pdbbase;

     indexDbBase = getIndexDbBase(env, obj);
     pdbbase = papdbBase[indexDbBase];
     str = (*env)->GetStringUTFChars(env, menuName, 0);
     dbDumpMenu(pdbbase, (char *)str); 
}

JNIEXPORT void JNICALL Java_dbStatic_DbBase_dumpRecordType
  (JNIEnv *env, jobject obj, jstring recordTypeName)
{
     jint     indexDbBase = -1;
     const    char *str;
     const    char *tstr;
     DBBASE    *pdbbase;

     indexDbBase = getIndexDbBase(env, obj);
     pdbbase = papdbBase[indexDbBase]; 
     str = (*env)->GetStringUTFChars(env, recordTypeName, 0);
     tstr = ((str==0) || (strlen(str)==0)) ? 0 : str;

     dbDumpRecordType(pdbbase, (char *)tstr);
}

JNIEXPORT void JNICALL Java_dbStatic_DbBase_dumpFldDes
  (JNIEnv *env, jobject obj, jstring recordTypeName, jstring fname)
{
  
     jint     indexDbBase = -1;
     const    char *strRecord;
     const    char *strFname;
     DBBASE    *pdbbase;

     indexDbBase = getIndexDbBase(env, obj);
     pdbbase = papdbBase[indexDbBase];
     strRecord = (*env)->GetStringUTFChars(env, recordTypeName, 0);
     strFname = (*env)->GetStringUTFChars(env, fname, 0);
     dbDumpFldDes(pdbbase, (char *)strRecord, (char *)strFname);
}
 
JNIEXPORT void JNICALL Java_dbStatic_DbBase_dumpDevice
  (JNIEnv *env, jobject obj, jstring recordTypeName)
{
     jint   indexDbBase = -1;
     const   char *str;
     DBBASE  *pdbbase;

     indexDbBase = getIndexDbBase(env, obj);
     pdbbase = papdbBase[indexDbBase];
     str = (*env)->GetStringUTFChars(env, recordTypeName, 0);
     dbDumpDevice(pdbbase, (char *)str);
} 

JNIEXPORT void JNICALL Java_dbStatic_DbBase_dumpDriver
  (JNIEnv *env, jobject obj)
{

     jint   indexDbBase = -1;
     const   char *str;
     DBBASE  *pdbbase;
 
     indexDbBase = getIndexDbBase(env, obj);
     pdbbase = papdbBase[indexDbBase];
     dbDumpDriver(pdbbase);
}

JNIEXPORT void JNICALL Java_dbStatic_DbBase_dumpBreaktable
  (JNIEnv *env, jobject obj, jstring name)
{
     jint   indexDbBase = -1;
     const   char *str;
     DBBASE  *pdbbase;
 
     indexDbBase = getIndexDbBase(env, obj);
     pdbbase = papdbBase[indexDbBase];
     str = (*env)->GetStringUTFChars(env, name, 0);
     dbDumpBreaktable(pdbbase, (char *)str);
}

JNIEXPORT void JNICALL Java_dbStatic_DbBase_pvdDump
  (JNIEnv *env, jobject obj, jint verbose)
{
     jint   indexDbBase = -1;
     DBBASE  *pdbbase;
 
     indexDbBase = getIndexDbBase(env, obj);
     pdbbase = papdbBase[indexDbBase];
     dbPvdDump(pdbbase, verbose);
}

JNIEXPORT void JNICALL Java_dbStatic_DbBase_reportDeviceConfig
  (JNIEnv *env, jobject obj, jstring report)
{
     jint   indexDbBase = -1;
     DBBASE  *pdbbase;
     const   char *str;
     FILE    *fd;
 
     indexDbBase = getIndexDbBase(env, obj);
     pdbbase = papdbBase[indexDbBase];
     str = (*env)->GetStringUTFChars(env, report, 0);
     fd = fopen(str, "w");
     dbReportDeviceConfig(pdbbase, (FILE *)fd);     
     dbReportDeviceConfig(pdbbase, stdout);  
     fclose(fd);
}


 
